(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('redux'), require('memoizerific'), require('immer'), require('redux-thunk'), require('react')) :
  typeof define === 'function' && define.amd ? define(['exports', 'redux', 'memoizerific', 'immer', 'redux-thunk', 'react'], factory) :
  (global = global || self, factory(global.EasyPeasy = {}, global.Redux, global.memoizerific, global.immer, global.ReduxThunk, global.React));
}(this, function (exports, redux, memoizerific, produce, reduxThunk, React) { 'use strict';

  memoizerific = memoizerific && memoizerific.hasOwnProperty('default') ? memoizerific['default'] : memoizerific;
  var produce__default = 'default' in produce ? produce['default'] : produce;
  reduxThunk = reduxThunk && reduxThunk.hasOwnProperty('default') ? reduxThunk['default'] : reduxThunk;
  var React__default = 'default' in React ? React['default'] : React;

  function _extends() {
    _extends = Object.assign || function (target) {
      for (var i = 1; i < arguments.length; i++) {
        var source = arguments[i];

        for (var key in source) {
          if (Object.prototype.hasOwnProperty.call(source, key)) {
            target[key] = source[key];
          }
        }
      }

      return target;
    };

    return _extends.apply(this, arguments);
  }

  var isStateObject = function isStateObject(x) {
    return x !== null && typeof x === 'object' && !Array.isArray(x);
  };

  /**
   * immer is an implementation detail, so we are not going to use its auto freeze
   * behaviour, which throws errors if trying to mutate state. It's also risky
   * for production builds as has a perf overhead.
   *
   * @see https://github.com/mweststrate/immer#auto-freezing
   */

  produce.setAutoFreeze(false);
  var maxSelectFnMemoize = 100;
  var actionNameSymbol = '__actionName__';
  var effectSymbol = '__effect__';
  var thunkSymbol = '__thunk__';
  var listenersSymbol = '__listeners__';
  var listenSymbol = '__listen__';
  var metaSymbol = '__meta__';
  var selectSymbol = '__select__';
  var selectImpSymbol = '__selectImp__';
  var selectDependenciesSymbol = '__selectDependencies__';
  var selectStateSymbol = '__selectState__';
  var reducerSymbol = '__reducer__';

  var get = function get(path, target) {
    return path.reduce(function (acc, cur) {
      return isStateObject(acc) ? acc[cur] : undefined;
    }, target);
  };

  var set$1 = function set(path, target, value) {
    path.reduce(function (acc, cur, idx) {
      if (idx + 1 === path.length) {
        acc[cur] = value;
      } else {
        acc[cur] = acc[cur] || {};
      }

      return acc[cur];
    }, target);
  };

  var tick = function tick() {
    return new Promise(function (resolve) {
      return setTimeout(resolve);
    });
  };

  var startsWith = function startsWith(target, search) {
    return target.substr(0, search.length) === search;
  };

  var notifiedAboutEffectDeprecation = false;
  var effect = function effect(fn) {
    if (!notifiedAboutEffectDeprecation) {
      notifiedAboutEffectDeprecation = true; // eslint-disable-next-line no-console

      console.warn('Easy Peasy: warning we have deprecated the "effect" helper. It will be removed in the next major release. We recommend that you migrate to the "thunk" helper, which satisfies the same responsibilities.');
    }

    fn[effectSymbol] = true;
    return fn;
  };
  var thunk = function thunk(fn) {
    fn[thunkSymbol] = true;
    return fn;
  };
  var select = function select(fn, dependencies) {
    fn[selectSymbol] = true;
    fn[selectDependenciesSymbol] = dependencies;
    fn[selectStateSymbol] = {};
    return fn;
  };
  var reducer = function reducer(fn) {
    fn[reducerSymbol] = true;
    return fn;
  };
  var notifiedAboutListenersDeprecation = false;
  var listeners = function listeners(fn) {
    if (!notifiedAboutListenersDeprecation) {
      notifiedAboutListenersDeprecation = true; // eslint-disable-next-line no-console

      console.warn('Easy Peasy: warning we have deprecated the "listeners" helper. It will be removed in the next major release. We recommend that you migrate to the "listen" helper, which satisfies the same responsibilities.');
    }

    fn[listenersSymbol] = true;
    return fn;
  };
  var listen = function listen(fn) {
    fn[listenSymbol] = true;
    return fn;
  };
  var createStore = function createStore(model, options) {
    if (options === void 0) {
      options = {};
    }

    var _options = options,
        _options$devTools = _options.devTools,
        devTools = _options$devTools === void 0 ? true : _options$devTools,
        _options$disableInter = _options.disableInternalSelectFnMemoize,
        disableInternalSelectFnMemoize = _options$disableInter === void 0 ? false : _options$disableInter,
        _options$middleware = _options.middleware,
        middleware = _options$middleware === void 0 ? [] : _options$middleware,
        _options$initialState = _options.initialState,
        initialState = _options$initialState === void 0 ? {} : _options$initialState,
        injections = _options.injections,
        compose = _options.compose,
        _options$reducerEnhan = _options.reducerEnhancer,
        reducerEnhancer = _options$reducerEnhan === void 0 ? function (rootReducer) {
      return rootReducer;
    } : _options$reducerEnhan;

    var wrapFnWithMemoize = function wrapFnWithMemoize(x) {
      return !disableInternalSelectFnMemoize && typeof x === 'function' ? memoizerific(maxSelectFnMemoize)(x) : x;
    };

    var definition = _extends({}, model, {
      logFullState: function logFullState(state) {
        // eslint-disable-next-line no-console
        console.log(JSON.stringify(state, null, 2));
      }
    });

    var references = {};
    var defaultState = {};
    var actionThunks = {};
    var actionCreators = {};
    var actionCreatorDict = {};
    var actionReducers = {};
    var customReducers = [];
    var selectorReducers = [];
    var listenerDefinitions = [];
    var listenerDict = {};
    var listenDefinitions = [];
    var listenDict = {};
    var listenStringDict = {};

    var dispatchListenersForAction = function dispatchListenersForAction(actionName, payload) {
      var listenersForAction = listenerDict[actionName];
      var listensForAction = [].concat(listenDict[actionName] || [], listenStringDict[actionName] || []);
      return Promise.all([listenersForAction && listenersForAction.length > 0 ? Promise.all(listenersForAction.map(function (listenerForAction) {
        return listenerForAction(references.dispatch, payload, references.getState, injections);
      })) : Promise.resolve(), listensForAction && listensForAction.length > 0 ? Promise.all(listensForAction.map(function (listenForAction) {
        return listenForAction(get(listenForAction[metaSymbol].parent, actionCreators), payload, {
          dispatch: references.dispatch,
          getState: references.getState,
          injections: injections,
          meta: listenForAction[metaSymbol]
        });
      })) : Promise.resolve()]);
    };

    var extract = function extract(current, parentPath) {
      return Object.keys(current).forEach(function (key) {
        var value = current[key];
        var path = [].concat(parentPath, [key]);
        var meta = {
          parent: parentPath,
          path: path
        };

        if (typeof value === 'function') {
          if (value[selectSymbol]) {
            // skip
            value[selectStateSymbol] = {
              parentPath: parentPath,
              key: key,
              executed: false
            };
            selectorReducers.push(value);
          } else if (value[thunkSymbol]) {
            var actionName = "@thunk." + path.join('.');
            value[actionNameSymbol] = actionName; // Thunk Action

            var action = function action(payload) {
              return value(get(parentPath, actionCreators), payload, {
                dispatch: references.dispatch,
                getState: references.getState,
                injections: injections,
                meta: meta
              });
            };

            set$1(path, actionThunks, action); // Thunk Action Creator

            var actionCreator = function actionCreator(payload) {
              return tick().then(function () {
                return references.dispatch({
                  type: actionName + "(started)",
                  payload: payload
                });
              }).then(function () {
                return references.dispatch(function () {
                  return action(payload);
                });
              }).then(function (result) {
                references.dispatch({
                  type: actionName + "(completed)",
                  payload: payload
                });
                return result;
              });
            };

            actionCreator[actionNameSymbol] = actionName;
            actionCreatorDict[actionName] = actionCreator;
            set$1(path, actionCreators, actionCreator);
          } else if (value[effectSymbol]) {
            var _actionName = "@effect." + path.join('.');

            value[actionNameSymbol] = _actionName; // Effect Action

            var _action = function _action(payload) {
              if (devTools) {
                references.dispatch({
                  type: _actionName,
                  payload: payload
                });
              }

              return value(references.dispatch, payload, references.getState, injections, {
                parent: parentPath,
                path: path
              });
            }; // Effect Action Creator


            var _actionCreator = function _actionCreator(payload) {
              return tick().then(function () {
                return references.dispatch(function () {
                  return _action(payload);
                });
              }).then(function (result) {
                dispatchListenersForAction(_actionCreator[actionNameSymbol], payload);
                return result;
              });
            };

            _actionCreator[actionNameSymbol] = _actionName;
            actionCreatorDict[_actionName] = _actionCreator;
            set$1(path, actionCreators, _actionCreator);
          } else if (value[reducerSymbol]) {
            customReducers.push({
              path: path,
              reducer: value
            });
          } else if (value[listenersSymbol]) {
            listenerDefinitions.push(value);
          } else if (value[listenSymbol]) {
            listenDefinitions.push(value);
            value[metaSymbol] = meta;
          } else {
            var _actionName2 = "@action." + path.join('.');

            value[actionNameSymbol] = _actionName2; // Reducer Action

            var _action2 = function _action2(state, payload) {
              return produce__default(state, function (draft) {
                return value(draft, payload, {
                  dispatch: references.dispatch,
                  dispatchLocal: get(path, references.dispatch),
                  getState: references.getState
                });
              });
            };

            _action2.actionName = _actionName2;
            set$1(path, actionReducers, _action2); // Reducer Action Creator

            var _actionCreator2 = function _actionCreator2(payload) {
              var result = references.dispatch({
                type: _action2.actionName,
                payload: payload
              });
              dispatchListenersForAction(_actionCreator2[actionNameSymbol], payload);
              return result;
            };

            _actionCreator2[actionNameSymbol] = _actionName2;
            actionCreatorDict[_actionName2] = _actionCreator2;
            set$1(path, actionCreators, _actionCreator2);
          }
        } else if (isStateObject(value) && Object.keys(value).length > 0) {
          extract(value, path);
        } else {
          // State
          var initialParentRef = get(parentPath, initialState);

          if (initialParentRef && key in initialParentRef) {
            set$1(path, defaultState, initialParentRef[key]);
          } else {
            set$1(path, defaultState, value);
          }
        }
      });
    };

    extract(definition, []);
    selectorReducers.forEach(function (selector) {
      selector[selectImpSymbol] = memoizerific(1)(function (state) {
        return wrapFnWithMemoize(selector(state));
      });
    });
    listenDefinitions.forEach(function (def) {
      var on = function on(action, thunkHandler) {
        if (typeof thunkHandler !== 'function') {
          return;
        }

        thunkHandler[metaSymbol] = def[metaSymbol];

        if (typeof action === 'function' && action[actionNameSymbol] && actionCreatorDict[action[actionNameSymbol]]) {
          listenDict[action[actionNameSymbol]] = listenDict[action[actionNameSymbol]] || [];
          listenDict[action[actionNameSymbol]].push(thunkHandler);
        } else if (typeof action === 'string') {
          listenStringDict[action] = listenStringDict[action] || [];
          listenStringDict[action].push(thunkHandler);
        }
      };

      def(on);
    });
    listenerDefinitions.forEach(function (def) {
      var on = function on(actionCreator, handler) {
        if (typeof actionCreator === 'function' && actionCreator[actionNameSymbol] && actionCreatorDict[actionCreator[actionNameSymbol]]) {
          listenerDict[actionCreator[actionNameSymbol]] = listenerDict[actionCreator[actionNameSymbol]] || [];
          listenerDict[actionCreator[actionNameSymbol]].push(handler);
        }
      };

      def(actionCreators, on);
    });

    var createReducers = function createReducers() {
      var createActionsReducer = function createActionsReducer(current, path) {
        var actionReducersAtPath = Object.keys(current).reduce(function (acc, key) {
          var value = current[key];

          if (typeof value === 'function' && !value[effectSymbol]) {
            return [].concat(acc, [value]);
          }

          return acc;
        }, []);
        var stateAtPath = Object.keys(current).reduce(function (acc, key) {
          return isStateObject(current[key]) ? [].concat(acc, [key]) : acc;
        }, []);
        var nestedReducers = stateAtPath.map(function (key) {
          return [key, createActionsReducer(current[key], [].concat(path, [key]))];
        });
        return function (state, action) {
          if (state === void 0) {
            state = get(path, defaultState);
          }

          // short circuit effects as they are noop in reducers
          if (startsWith(action.type, '@effect.')) {
            return state;
          } // short circuit actions if they aren't a match on current path


          if (path.length > 0 && !startsWith(action.type, "@action." + path.join('.'))) {
            return state;
          }

          var actionReducer = actionReducersAtPath.find(function (x) {
            return x.actionName === action.type;
          });

          if (actionReducer) {
            return actionReducer(state, action.payload);
          }

          for (var i = 0; i < nestedReducers.length; i += 1) {
            var _nestedReducers$i = nestedReducers[i],
                key = _nestedReducers$i[0],
                red = _nestedReducers$i[1];
            var newState = red(state[key], action);

            if (state[key] !== newState) {
              var _extends2;

              return _extends({}, state, (_extends2 = {}, _extends2[key] = newState, _extends2));
            }
          }

          return state;
        };
      };

      var reducerForActions = createActionsReducer(actionReducers, []);
      var reducerWithCustom = customReducers.length > 0 ? function (state, action) {
        var stateAfterActions = reducerForActions(state, action);
        return produce__default(stateAfterActions, function (draft) {
          customReducers.forEach(function (_ref) {
            var p = _ref.path,
                red = _ref.reducer;
            var current = get(p, draft);
            set$1(p, draft, red(current, action));
          });
        });
      } : reducerForActions;

      var runSelectorReducer = function runSelectorReducer(state, selector) {
        var _selector$selectState = selector[selectStateSymbol],
            parentPath = _selector$selectState.parentPath,
            key = _selector$selectState.key,
            executed = _selector$selectState.executed;

        if (executed) {
          return state;
        }

        var dependencies = selector[selectDependenciesSymbol];
        var stateAfterDependencies = dependencies ? dependencies.reduce(runSelectorReducer, state) : state;
        var newState = state;

        if (parentPath.length > 0) {
          var target = get(parentPath, stateAfterDependencies);

          if (target) {
            var newValue = selector[selectImpSymbol](target);
            newState = produce__default(state, function (draft) {
              var updateTarget = get(parentPath, draft);
              updateTarget[key] = newValue;
            });
          }
        } else {
          var _newValue = selector[selectImpSymbol](stateAfterDependencies);

          newState = produce__default(state, function (draft) {
            draft[key] = _newValue;
          });
        }

        selector[selectStateSymbol].executed = true;
        return newState;
      };

      var runSelectors = function runSelectors(state) {
        return selectorReducers.reduce(runSelectorReducer, state);
      };

      var isInitial = true;
      return selectorReducers.length > 0 ? function (state, action) {
        var stateAfterActions = reducerWithCustom(state, action);

        if (state !== stateAfterActions || isInitial) {
          var stateAfterSelectors = runSelectors(stateAfterActions);
          isInitial = false;
          selectorReducers.forEach(function (selector) {
            selector[selectStateSymbol].executed = false;
          });
          return stateAfterSelectors;
        }

        return stateAfterActions;
      } : reducerWithCustom;
    };

    var reducers = createReducers();
    var composeEnhancers = compose || (devTools && typeof window !== 'undefined' && window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ ? window.__REDUX_DEVTOOLS_EXTENSION_COMPOSE__ : redux.compose);

    var dispatchActionStringListeners = function dispatchActionStringListeners() {
      return function (next) {
        return function (action) {
          if (listenStringDict[action.type]) {
            dispatchListenersForAction(action.type, action.payload);
          }

          next(action);
        };
      };
    };

    var store = redux.createStore(reducerEnhancer(reducers), defaultState, composeEnhancers(redux.applyMiddleware.apply(void 0, [reduxThunk, dispatchActionStringListeners].concat(middleware)))); // attach the action creators to dispatch

    Object.keys(actionCreators).forEach(function (key) {
      store.dispatch[key] = actionCreators[key];
    });
    references.dispatch = store.dispatch;
    references.getState = store.getState;
    references.getState.getState = store.getState;
    return store;
  };

  //

  var shallowequal = function shallowEqual(objA, objB, compare, compareContext) {
    var ret = compare ? compare.call(compareContext, objA, objB) : void 0;

    if (ret !== void 0) {
      return !!ret;
    }

    if (objA === objB) {
      return true;
    }

    if (typeof objA !== "object" || !objA || typeof objB !== "object" || !objB) {
      return false;
    }

    var keysA = Object.keys(objA);
    var keysB = Object.keys(objB);

    if (keysA.length !== keysB.length) {
      return false;
    }

    var bHasOwnProperty = Object.prototype.hasOwnProperty.bind(objB);

    // Test for A's keys different from B.
    for (var idx = 0; idx < keysA.length; idx++) {
      var key = keysA[idx];

      if (!bHasOwnProperty(key)) {
        return false;
      }

      var valueA = objA[key];
      var valueB = objB[key];

      ret = compare ? compare.call(compareContext, valueA, valueB, key) : void 0;

      if (ret === false || (ret === void 0 && valueA !== valueB)) {
        return false;
      }
    }

    return true;
  };

  var StoreContext = React.createContext();

  function useStore(mapState, dependencies) {
    if (dependencies === void 0) {
      dependencies = [];
    }

    var store = React.useContext(StoreContext);

    var _useState = React.useState(mapState(store.getState())),
        state = _useState[0],
        setState = _useState[1]; // As our effect only fires on mount and unmount it won't have the state
    // changes visible to it, therefore we use a mutable ref to track this.


    var stateRef = React.useRef(state); // Helps avoid firing of events when unsubscribed, i.e. unmounted

    var isActive = React.useRef(true);
    React.useEffect(function () {
      var calculateState = function calculateState() {
        var newState = mapState(store.getState());
        isActive.current = true;

        if (newState === stateRef.current || isStateObject(newState) && isStateObject(stateRef.current) && shallowequal(newState, stateRef.current)) {
          // Do nothing
          return;
        }

        stateRef.current = newState; // The settimeout wrap fixes a strange issue where a setState would
        // fire but the associated hook wouldn't receive it. It's almost as
        // if the effect was handled in a synchronous manner in some part of
        // the React reconciliation process that ended up with it not
        // propagating

        setTimeout(function () {
          if (isActive.current) {
            setState(newState);
          }
        });
      };

      calculateState();
      var unsubscribe = store.subscribe(calculateState);
      return function () {
        unsubscribe();
        isActive.current = false;
      };
    }, dependencies);
    return state;
  }
  var warnedAboutUseActionDeprecation = false;
  function useAction(mapActions) {
    if (!warnedAboutUseActionDeprecation) {
      warnedAboutUseActionDeprecation = true; // eslint-disable-next-line no-console

      console.warn('Easy Peasy: the "useAction" hook has been deprecated and will be removed in the next major release. We recommend using the "useActions" hook instead.');
    }

    var store = React.useContext(StoreContext);
    return mapActions(store.dispatch);
  }
  function useActions(mapActions) {
    var store = React.useContext(StoreContext);
    return mapActions(store.dispatch);
  }
  function useDispatch() {
    var store = React.useContext(StoreContext);
    return store.dispatch;
  }
  function createTypedHooks() {
    return {
      useActions: useActions,
      useAction: useAction,
      useDispatch: useDispatch,
      useStore: useStore
    };
  }

  var StoreProvider = function StoreProvider(_ref) {
    var children = _ref.children,
        store = _ref.store;
    return React__default.createElement(StoreContext.Provider, {
      value: store
    }, children);
  };

  exports.createStore = createStore;
  exports.createTypedHooks = createTypedHooks;
  exports.effect = effect;
  exports.listeners = listeners;
  exports.listen = listen;
  exports.reducer = reducer;
  exports.select = select;
  exports.StoreProvider = StoreProvider;
  exports.thunk = thunk;
  exports.useAction = useAction;
  exports.useActions = useActions;
  exports.useDispatch = useDispatch;
  exports.useStore = useStore;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=easy-peasy.js.map
